
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#ifndef _MOOF_PLANE_HH_
#define _MOOF_PLANE_HH_

#include <Moof/Math.hh>
#include <Moof/Shape.hh>


namespace Mf {


template <int D> class Aabb;
template <int D> class Sphere;


/*
 * A plane in 3-space defined by the equation Ax + By + Cz = D, where [A, B, C]
 * is normal to the plane.
 */

struct Plane : public Shape<3>
{
	Vector3	normal;
	Scalar	d;

	typedef enum
	{
		NEGATIVE	= -1,
		INTERSECT	=  0,
		POSITIVE	=  1
	} Halfspace;

	Plane() {}
	Plane(const Vector3& vector, Scalar scalar) :
		normal(vector),
		d(scalar) {}
	Plane(Scalar a, Scalar b, Scalar c, Scalar scalar) :
		normal(a, b, c),
		d(scalar) {}


	Scalar intersectRay(const Ray<3>& ray, Ray<3>::Intersection& intersection)
	{
		// solve: [(ray.point + t*ray.direction) dot normal] + d = 0

		Scalar denominator = cml::dot(ray.direction, normal);

		// check for parallel condition
		if (denominator == SCALAR(0.0))
		{
			if (isEqual(cml::dot(ray.point, normal), -d))
			{
				// the ray lies on the plane
				intersection.point = ray.point;
				intersection.normal.set(0.0, 0.0, 0.0);
				//intersection.normal = normal;
				return SCALAR(0.0);
			}

			// no solution
			return SCALAR(-1.0);
		}

		Scalar distance = cml::dot(ray.point, normal) + d;
		Scalar t = -distance / denominator;
		if (t > SCALAR(0.0))
		{
			ray.solve(intersection.point, t);

			if (distance >= 0.0) intersection.normal = normal;
			else                 intersection.normal = -normal;
		}

		return t;
	}


	/* Causes the normal of the plane to become normalized.  The scalar may also
	 * be changed to keep the equation true.  Word to the wise: don't normalize
	 * a plane if the normal is the zero vector. */
	void normalize()
	{
		Scalar mag = normal.length();

		normal /= mag;
		d /= mag;
	}

	/**
	 * Determine the shortest distance between a point and the plane. */

	Scalar getDistanceToPoint(const Vector3& point) const
	{
		return cml::dot(point, normal) + d;
	}

	Halfspace intersects(const Vector3& point) const
	{
		Scalar distance = getDistanceToPoint(point);

		if (isEqual(distance, 0.0)) return INTERSECT;
		else if (distance < 0.0)    return NEGATIVE;
		else                        return POSITIVE;
	}

	Halfspace intersects(const Aabb<3>& aabb) const;
	Halfspace intersects(const Sphere<3>& sphere) const;
};


} // namespace Mf

#endif // _MOOF_PLANE_HH_

/** vim: set ts=4 sw=4 tw=80: *************************************************/

